"""Class definition for GeneralPath object."""
from copy import deepcopy
from pathlib import Path
from urllib.parse import urlparse, urlunparse
from typing import List, Union, Tuple, Optional

import boto3


class GeneralPath:
    """Contains READ ONLY functionalities."""
    def __init__(self, uri: str):
        if '://' not in uri:
            raise ValueError('Invalid uri specified')
        self.url = urlparse(uri)
        self.path = Path(self.url.path.strip('/'))
        self.client = None

    def __str__(self) -> str:
        return urlunparse((self.url.scheme, self.url.netloc,
                           self.path.as_posix(), self.url.params,
                           self.url.query, self.url.fragment))

    def __repr__(self) -> str:
        return f"GeneralPath({str(self)})"

    def __copy__(self) -> 'GeneralPath':
        return self.__deepcopy__()

    def __deepcopy__(self, memodict={}) -> 'GeneralPath':
        return deepcopy(self)

    def __div__(self, other: str) -> 'GeneralPath':
        if not isinstance(other, str):
            raise ValueError(f'GeneralPath can only be append with str type')
        ret = self.copy()
        ret.path /= other
        return ret

    def _init_client(self) -> 'GeneralPath':
        if 's3' in self.url.scheme:
            self.client = S3Client(self.url.netloc)
        else:
            raise NotImplementedError()
        return self

    @property
    def name(self) -> str:
        return self.path.name

    @property
    def stem(self) -> str:
        return self.path.stem

    @property
    def suffix(self) -> str:
        return self.path.suffix

    def parts(self) -> Tuple[str]:
        return tuple([self.url.scheme] + self.path.parts)

    def as_uri(self) -> str:
        return str(self)

    def is_dir(self) -> bool:
        if self.client is None:
            self._init_client()
        result = self.client.search(str(self.path))
        if len(result) == 1 and result[0].endswith('/'):
            return True
        else:
            return False

    def is_file(self) -> bool:
        if self.client is None:
            self._init_client()
        result = self.client.search(str(self.path))
        if len(result) == 1 and '.' in self.path.name:
            return True
        else:
            return False

    def exists(self) -> bool:
        return self.is_file() or self.is_dir()

    def rglob(self, pattern: str = '') -> List['GeneralPath']:
        splits = pattern.split('*')
        if len(splits) > 2:
            raise ValueError('Only one wildcard character * allowed')

        if self.client is None:
            self._init_client()
        result = [GeneralPath(x) for x in self.client.search(splits[0])]

        if len(splits) == 2:
            result = [x for x in result if x.path.as_posix().endswith(splits[1])]

        return result

    def glob(self, pattern: str = '') -> List['GeneralPath']:
        ret = []
        base = pattern.split('*')[0]  # Bug when pattern is not a complete folder name
        for hit in self.rglob(pattern=pattern):
            if '/' not in str(hit.path)[(len(str(self.path)) + len(base) + 1):]:  # Plus one is to account for possible '/'
                ret.append(hit)
        return ret

    def download(self, dst: str) -> None:
        if self.is_file():
            Path(dst).parent.mkdir(parents=True, exist_ok=True)
            return self.client.download(str(self.path), dst)
        elif self.is_dir():
            raise NotImplementedError()
        else:
            raise FileNotFoundError('File does not exist')




class S3Client:
    def __init__(self, bucket: str) -> "S3Client":
        self.bucket = bucket.strip('/')
        self.client = boto3.client('s3')

    def __str__(self) -> str:
        pass

    def __repr__(self) -> str:
        pass

    def search(self, prefix: Optional[str] = None) -> List[str]:
        hits = []

        search_params = {'Bucket': self.bucket}
        if prefix not in (None, ''):
            search_params['Prefix'] = prefix

        response = self.client.list_objects_v2(**search_params)
        while True:
            if 'Contents' not in response:
                return []
            keys = [x['Key'] for x in response['Contents']]

            # Sort in increasing length
            keys = sorted(keys, key=len)

            # Only check if folders (ending with '/') is redundant
            ret = []
            for i, key in enumerate(keys):
                if not key.endswith('/'):
                    ret.append(key)
                    continue

                for j in range(i + 1, len(keys)):
                    if keys[i] in keys[j]:
                        break
                else:
                    ret.append(key)
            hits += ret

            # Continue search if 'NextContinuationToken' exists
            if 'NextContinuationToken' in response:
                search_params['ContinuationToken'] = response['NextContinuationToken']
                response = self.client.list_objects_v2(**search_params)
            else:
                break

        # Convert `hits` into S3 urls
        hits = [f's3://{self.bucket}/{x}' for x in hits]
        return hits

    def download(self, key: str, dst: str) -> None:
        self.client.download_file(Bucket=self.bucket, Key=key, Filename=dst)



from functools import partial
from typing import Optional, Dict, Any
from multiprocessing import cpu_count, Pool

from tqdm import tqdm


def parallelize(func,
                iterable,
                length: Optional[int] = None,
                n_workers: Optional[int] = None,
                desc: Optional[str] = None,
                chunksize: Optional[int] = None,
                async_ok: bool = False,
                leave: bool = True,
                **func_kwargs,
               ):
    workers = (cpu_count() - 1) if n_workers is None else n_workers
    chunksize = 1 if chunksize is None else chunksize
    total = len(iterable) if length is None else length
    func = partial(func, **func_kwargs)

    with Pool(workers) as p:
        imap_func = p.imap_unordered if async_ok else p.imap
        return list(tqdm(imap_func(func, iterable, chunksize=chunksize),
                         desc=desc, dynamic_ncols=True, leave=leave,
                         total=total))



import itertools as it
from typing import Any, Iterable, TypeVar


A = TypeVar('Any')


def batch(iterable: Iterable[A], size: int) -> Iterable[A]:
    iters = iter(iterable)
    def take():
        while True:
            yield tuple(it.islice(iters, size))
    return iter(take().__next__, tuple())



"""
This module contains similarity- and diversity-based data selection metrics.

These metrics were proposed by Ruder and Plank
in their `paper <https://arxiv.org/pdf/1707.05246.pdf>`_.
"""

from .similarity import SIMILARITY_FEATURES, similarity_func_factory
from .diversity import DIVERSITY_FEATURES, diversity_func_factory



import hashlib


def md5(fname: str, buffer_size: int = 4096) -> str:
    md5_hash = hashlib.md5()
    with open(fname, 'rb') as f:
        for chunk in iter(lambda: f.read(buffer_size), b''):
            md5_hash.update(chunk)
    return md5_hash.hexdigest()



from collections import Counter
from functools import reduce
from typing import List, Optional, Sequence, Set, Union, Counter as CounterType

import numpy as np
import pandas as pd
from tqdm import tqdm
from scipy import sparse
from sklearn.preprocessing import RobustScaler
from transformers import PreTrainedTokenizerFast
from sklearn.base import BaseEstimator, TransformerMixin

from transformers_domain_adaptation.type import Corpus, Token
from transformers_domain_adaptation.data_selection.metrics import (
    SIMILARITY_FEATURES,
    DIVERSITY_FEATURES,
    similarity_func_factory,
    diversity_func_factory,
)


class DataSelector(BaseEstimator, TransformerMixin):
    """Select subset of data that is likely to be beneficial for domain pre-training.

    This class is sklearn-compatible and implements the sklearn Transformers interface.
    """

    def __init__(
        self,
        keep: Union[int, float],
        tokenizer: PreTrainedTokenizerFast,
        similarity_metrics: Optional[Sequence[str]] = None,
        diversity_metrics: Optional[Sequence[str]] = None,
    ):
        """
        Args:
            keep: Quantity of documents from corpus to keep.
                  To specify number of documents, use :obj:`int`.
                  To specify percentage of documents in corpus, use :obj:`float`.
            tokenizer: A Rust-based 🤗 Tokenizer
            similarity_metrics: An optional list of similarity metrics
            diversity_metrics: An optional list of diversity metrics

        Note:
            For a list of similarity and diversity metrics, refer to :ref:`data-selection-metrics`

        Note:
            At least one similarity/diversity metric must be provided.
        """
        if isinstance(keep, int) and keep <= 0:
            raise ValueError(f"Int value for `keep` must be strictly positive.")
        if isinstance(keep, float) and not 0 <= keep <= 1:
            raise ValueError(
                f"Float value for `keep` must be between 0 and 1 (inclusive)."
            )
        if similarity_metrics is not None:
            _invalid_sim_metrics = set(similarity_metrics) - SIMILARITY_FEATURES
            if _invalid_sim_metrics:
                raise ValueError(
                    f"Invalid similarity metric(s) {_invalid_sim_metrics} found"
                )
        if diversity_metrics is not None:
            _invalid_div_metrics = set(diversity_metrics) - DIVERSITY_FEATURES
            if _invalid_div_metrics:
                raise ValueError(
                    f"Invalid diversity metric(s) {_invalid_div_metrics} found"
                )
        if similarity_metrics is None and diversity_metrics is None:
            raise ValueError(
                f"No metrics provided. Please provide at least one similarity or diversity metric."
            )

        self.keep = keep
        self.tokenizer = tokenizer
        self.similarity_metrics = similarity_metrics
        self.diversity_metrics = diversity_metrics

    def _to_term_counts(self, texts: Sequence[str]) -> List[CounterType[str]]:
        # Tokenize all documents using Rust tokenizer
        counters: List[CounterType[str]] = [
            Counter(enc.tokens)
            for enc in self.tokenizer.backend_tokenizer.encode_batch(
                texts, add_special_tokens=False
            )
        ]
        return counters

    @staticmethod
    def _to_term_dists(counters: List[CounterType[str]]) -> sparse.csr_matrix:
        vocab: Set[str] = reduce(lambda x, y: x | set(y), counters, set())
        vocab_mapping = {v: i for i, v in enumerate(vocab)}

        rows = np.array([val for i, counter in enumerate(counters) for val in [i] * len(counter)])
        cols = np.array([
            vocab_id for counter in counters for vocab_id in map(vocab_mapping.get, counter)
        ])
        data = []
        for counter in counters:
            n_elems = sum(counter.values())
            data.extend([val / n_elems for val in counter.values()])

        term_dists = sparse.csr_matrix(
            (data, (rows, cols)),
            shape=(len(counters), len(vocab)),
            dtype=np.float64,
        )
        return term_dists

    def fit(self, ft_corpus: Corpus):
        """Compute corpus-level term distribution of :obj:`ft_corpus`.

        Args:
            ft_corpus: The fine-tuning corpus. Not to be confused with
                       the domain pre-training corpus (which is used in :meth:`transform`)

        Note:
            The :obj:`ft_corpus` is treated as a single "document", which will be compared
            against other documents in the in-domain corpus in :meth:`transform`
        """
        # We cannot obtain a term distribution until we obtain the entire vocabulary of texts in `.transform`
        self.ft_term_counts_ = self._to_term_counts([' '.join(ft_corpus)])[0]
        return self

    def transform(self, docs: Corpus) -> Corpus:
        """Create a relevant subset of documents from the training corpus based on the provided data selection metrics.

        Args:
            docs: The training corpus

        Returns:
            A subset of relevant :obj:`docs` for domain pre-training
        """
        docs = [doc for doc in docs if len(doc.strip()) > 1]
        scores = self.compute_metrics(docs)
        composite_scores = scores["composite"].sort_values(ascending=False)

        n_select = (
            self.keep if isinstance(self.keep, int) else int(self.keep * len(docs))
        )
        selection_index = composite_scores.index[:n_select]
        subset_corpus = pd.Series(docs)[selection_index]

        return subset_corpus.tolist()

    def compute_metrics(self, docs: Corpus) -> pd.DataFrame:
        scores = pd.concat(
            [
                self.compute_similarities(docs),
                self.compute_diversities(docs),
            ],
            axis=1,
        )

        # Ensure metrics are normalized, before combining them into a composite score
        scores = pd.DataFrame(
            RobustScaler().fit_transform(scores), columns=scores.columns
        )
        scores["composite"] = scores.sum(axis=1)
        return scores

    def compute_similarities(self, docs: Corpus) -> pd.DataFrame:
        similarities = pd.DataFrame()  # of shape (n_docs, n_metrics)
        if (
            self.similarity_metrics is None
        ):  # Short-circuit function to avoid unnecessary computations
            return similarities

        term_counts = self._to_term_counts(docs)
        _term_dists = self._to_term_dists([self.ft_term_counts_] + term_counts)
        self.ft_term_dist_, term_dists = _term_dists[0], _term_dists[1:]

        pbar = tqdm(
            self.similarity_metrics,
            desc="computing similarity",
            unit="metric",
            dynamic_ncols=True,
        )
        for metric in pbar:
            sim_func = similarity_func_factory(metric)
            similarities[metric] = sim_func(
                term_dists, self.ft_term_dist_.toarray().reshape(1, -1)
            )

        return similarities

    def compute_diversities(self, docs: Corpus) -> pd.DataFrame:
        diversities = pd.DataFrame()  # of shape (n_docs, n_metrics)
        if self.diversity_metrics is None:
            return diversities

        tokenized_docs: List[List[Token]] = [
            enc.tokens for enc in self.tokenizer.backend_tokenizer.encode_batch(docs)
        ]

        pbar = tqdm(
            self.diversity_metrics,
            desc="computing diversity",
            unit="metric",
            dynamic_ncols=True,
        )
        for metric in pbar:
            div_func = diversity_func_factory(
                metric,
                train_term_dist=self.ft_term_dist_,
                vocab2id=self.tokenizer.vocab,
            )
            diversities[metric] = pd.Series(
                (div_func(tokenized_doc) for tokenized_doc in tokenized_docs)
            )

        return diversities



"""Script to shard large text file.

Borrowed from https://gist.github.com/Guitaricet/937264d4958cb62084008cc5d6da0685
"""
import shutil
import logging
import argparse
from pathlib import Path

from tqdm import tqdm


logger = logging.getLogger(__name__)


def parse_args():
    parser = argparse.ArgumentParser(
        'General script to shard a text file.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument('--src', type=Path, required=True,
                        help='Input text file to be sharded')
    parser.add_argument('--dst', type=Path, required=True,
                        help='Folder to store sharded files')
    parser.add_argument('-s', '--shard-size', default=int(1e6), type=int,
                        help='Number of lines per shard')
    return parser.parse_args()


def read_in_chunks(file_object, n_lines=1024) -> str:
    while True:
        ret = []
        for _ in range(n_lines):
            data = file_object.readline()
            if not data:
                break
            ret.append(data)

        if not len(ret):
            break
        yield ''.join(ret)


def main(args):
    logger.info(f'Writing shards to {args.dst}...')
    if args.dst.exists():
        logger.info(f'Removed existing folder at {args.dst}')
        shutil.rmtree(str(args.dst), ignore_errors=True)
    args.dst.mkdir(parents=True)

    with open(args.src) as f:
        for i, chunk in tqdm(enumerate(read_in_chunks(f,
                                                      n_lines=args.shard_size)),
                             desc='Sharding'):
            shard = (
                (args.dst / f'{args.src.stem}_{i:05d}')
                .with_suffix('.txt')
            )
            tqdm.write(f'Created shard {shard}')
            shard.write_text(chunk)

    logger.info(f'Successfully sharded {args.src} into {i} shards at {args.dst}')


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    args = parse_args()
    main(args)



"""Text-related util functions."""
from ftfy import fix_text


def clean(text: str, lowercase: bool = False) -> str:
    """Clean text.

    Does the following:
        1. Replace whitespace characters with space
        2. Fix corrupted text encodings
        3. Lowercase (if specified)
    
    Arguments:
        text {str} -- Text to clean
    
    Keyword Arguments:
        lowercase {bool} -- If True, perform lowercasing (default: {False})
    
    Returns:
        str -- Cleaned text
    """
    # Replace whitespace characters with spaces
    text = ' '.join(text.split())

    # Fix encoded texts
    text = fix_text(text)

    # Lowercase text if specified
    if lowercase:
        text = text.lower()

    return text



from typing import List
import urllib.request

import requests
from bs4 import BeautifulSoup
from bs4.dammit import EncodingDetector


def find_hlinks(endpoint: str) -> List[str]:
    resp = requests.get(endpoint)

    # Find appropriate encodings
    http_enc = resp.encoding if 'chatset' in resp.headers.get('content-type', '').lower() else None
    html_enc = EncodingDetector.from_declared_encoding(resp.content, is_html=True)
    enc = html_enc or http_enc

    soup = BeautifulSoup(resp.content, from_encoding=encoding)

    return [link['href'] for link in soup.find_all('a', href=True)]


def download_file(url: str, dst: str) -> None:
    urllib.request.urlretrieve(url, dst)



from pathlib import Path
from setuptools import setup, find_packages


HERE = Path(__file__).parent
README = (HERE / "README.md").read_text()
with open("README.md", encoding="utf-8") as f:
    LONG_DESCRIPTION = f.read()

setup(
    name="transformers-domain-adaptation",
    version="0.3.0a10",
    description="Adapt Transformer-based language models to new text domains",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/georgianpartners/Transformers-Domain-Adaptation",
    author="Christopher Tee",
    author_email="chris@georgian.io",
    license="MIT",
    python_requires=">=3.6",
    keywords=[
        "transformers",
        "tokenizers",
        "huggingface",
        "pytorch",
        "domain-adaptation",
        "transfer-learning",
        "natural-language-processing",
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Text Processing :: General",
    ],
    package_dir={"": "src"},
    packages=find_packages(
        where="src", exclude=("etl*", "utils*", "experimental*", "tests")
    ),
    install_requires=[
        "transformers>=4,<5",
        "tokenizers>=0.9,<0.10",
        "datasets>=1.2,<1.3",
        "pandas",
        "torch>=1.7,<1.8",
        "scipy==1.5.4",
        "scikit-learn",
        "tqdm",
    ],
)



"""Utility module for interacting with the shell."""
import shlex
from subprocess import run, Popen, PIPE


def run_shell(cmd: str) -> str:
    """Run a shell command using the subprocess module."""
    return run(shlex.split(cmd),
               stdout=PIPE, stderr=PIPE,
               check=True, universal_newlines=True)


def is_file_in_use(filename: str) -> bool:
    """Check if a file is being used by other processes."""
    lsout = Popen(shlex.split(f'lsof {filename}'), stdout=PIPE)
    output = run(shlex.split(f'grep {filename}'), stdin=lsout.stdout)
    return not output.returncode
